""" See joelnet/data.py by Joel Grus at
https://github.com/joelgrus/joelnet/blob/master/joelnet/data.py """

from typing import NamedTuple, Iterator

import numpy as np

from edugrad.tensor import Tensor

Batch = NamedTuple("Batch", [("inputs", Tensor), ("targets", Tensor)])


class DataIterator:
    def __call__(self, inputs: np.ndarray, targets: np.ndarray) -> Iterator[Batch]:
        raise NotImplementedError


class BatchIterator(DataIterator):
    def __init__(self, batch_size: int = 32, shuffle: bool = True) -> None:
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __call__(self, inputs: np.ndarray, targets: np.ndarray) -> Iterator[Batch]:
        starts = np.arange(0, len(inputs), self.batch_size)
        if self.shuffle:
            np.random.shuffle(starts)
        self.num_batches = len(starts)

        for start in starts:
            end = start + self.batch_size
            batch_inputs = Tensor(inputs[start:end], name="x")
            batch_targets = Tensor(targets[start:end], name="y")
            yield Batch(batch_inputs, batch_targets)



from typing import List, Callable

import numpy as np
from edugrad.tensor import Tensor


class Operation:
    @staticmethod
    def forward(
        ctx: List[np.ndarray], *inputs: List[np.ndarray], **kwargs
    ) -> np.ndarray:
        """Forward pass of an operation.

        Args:
            ctx: empty list of arrays; can be used to store values for backward pass
            inputs: arguments to this operation

        Returns:
            output of the operation, assumed to be one numpy array
        """
        raise NotImplementedError

    @staticmethod
    def backward(ctx: List[np.ndarray], grad_output: np.ndarray) -> List[np.ndarray]:
        """Backward pass of an op, returns dL / dx for each x in parents of this op.

        Args:
            ctx: stored values from the forward pass
            grad_output: dL/dv, where v is output of this node

        Returns:
            a _list_ of arrays, dL/dx, for each x that was input to this op
        """
        raise NotImplementedError


def tensor_op(op: Operation) -> Callable[[List[Tensor]], Tensor]:
    """
    Takes an operation, and turns it into a callable function on Tensors.

    The resulting function implicitly builds the dynamic computation graph,
    including populating the Tensors' _backward methods, when called.
    """

    def fn(*inputs: List[Tensor], **kwargs) -> Tensor:
        ctx = []
        new_tensor = Tensor(
            op.forward(ctx, *[tensor.value for tensor in inputs], **kwargs),
            inputs,
            op.__name__,
        )

        def _backward():
            grads = op.backward(ctx, new_tensor.grad)
            for idx in range(len(inputs)):
                inputs[idx].grad += grads[idx]

        new_tensor._backward = _backward
        return new_tensor

    return fn


@tensor_op
class add(Operation):
    def forward(ctx, a, b):
        return a + b

    def backward(ctx, grad_output):
        return grad_output, grad_output


@tensor_op
class minus(Operation):
    def forward(ctx, a, b):
        return a - b

    def backward(ctx, grad_output):
        return grad_output, -grad_output


@tensor_op
class matmul(Operation):
    def forward(ctx, mat1, mat2):
        ctx.append(mat1)
        ctx.append(mat2)
        return mat1 @ mat2

    def backward(ctx, grad_output):
        mat1, mat2 = ctx
        return grad_output @ mat2.T, mat1.T @ grad_output


@tensor_op
class power(Operation):
    """Raise to a power, e.g. a^exponent. """
    def forward(ctx, a, exponent = 1):
        ctx.append(a)
        ctx.append(exponent)
        return a ** exponent

    def backward(ctx, grad_output):
        value, exponent = ctx
        return [exponent * value ** (exponent-1) * grad_output]


@tensor_op
class relu(Operation):
    def forward(ctx, value):
        new_val = np.maximum(0, value)
        ctx.append(new_val)
        return new_val

    def backward(ctx, grad_output):
        value = ctx[-1]
        return [(value > 0).astype(float)]


@tensor_op
class reduce_sum(Operation):
    def forward(ctx, value):
        ctx.append(value)
        return np.sum(value)

    def backward(ctx, grad_output):
        return [np.ones(ctx[-1].shape) * grad_output]


@tensor_op
class reduce_mean(Operation):
    def forward(ctx, value):
        ctx.append(value)
        return np.mean(value)

    def backward(ctx, grad_output):
        shape = ctx[-1].shape
        return [np.ones(shape) * grad_output / np.prod(shape)]


@tensor_op
class copy_rows(Operation):
    """ Copies a (1, dim) array into (num, dim) """

    def forward(ctx, value, num=1):
        return np.tile(value, (num, 1))

    def backward(ctx, grad_output):
        return np.sum(grad_output, axis=0)


def mse_loss(predicted: Tensor, targets: Tensor) -> Tensor:
    """Computes mean( (yhat - y)^2 ) """
    diff = minus(predicted, targets)
    squared = diff**2
    return reduce_mean(squared)


""" Copyright 2020-2021 Shane Steinert-Threlkeld

    This file is part of edugrad.

    edugrad is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    edugrad is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with edugrad.  If not, see <https://www.gnu.org/licenses/>.
"""

""" An optimizer updates some parameters. """

from typing import Iterable

from edugrad.tensor import Tensor


class Optimizer:
    def step(self) -> None:
        raise NotImplementedError


class SGD(Optimizer):
    def __init__(self, params: Iterable[Tensor], lr=1e-4):
        self.lr = lr
        self._cur_step = 0
        self.params = params

    def step(self):
        for param in self.params:
            param.value -= self.lr * param.grad
        self._cur_step += 1



""" Copyright 2020-2021 Shane Steinert-Threlkeld

    This file is part of edugrad.

    edugrad is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    edugrad is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with edugrad.  If not, see <https://www.gnu.org/licenses/>.
"""

"""Defines the Module class, basic building block of neural networks.

Also includes a Linear Module, for computing xW + b.
"""

from typing import Callable, List
import numpy as np

from edugrad.tensor import Tensor
import edugrad.ops as ops


class Module:
    """Modules are components of computation graphs.

    They contain parameters (Tensors) and/or sub-modules.

    Most importantly, a Module has a `forward` method, which implements the
    forward pass of a computation graph.  This forward pass should use `ops`
    which ensures that the computation graph gets built dynamically, so that
    `backward` can be used to compute gradients.

    Attributes:
        _params: dictionary of Tensors, trainable Tensor parameters
        _modules: sub-modules of this module
    """
    def __init__(self):
        self._params = dict()
        self._modules = dict()

    def zero_grad(self) -> None:
        """Set gradients of all parameters to zero. """
        for param in self.parameters():
            param.grad = np.zeros(param.value.shape)

    def parameters(self) -> List[Tensor]:
        """Return a list of all parameters of this module. """
        params = list(self._params.values())
        for module in self._modules:
            params.extend(self._modules[module].parameters())
        return params

    def forward(self, *inputs: List[Tensor]) -> Tensor:
        raise NotImplementedError

    def __setattr__(self, key, value) -> None:
        """Register modules and params when assigned. """
        if isinstance(value, Tensor):
            self._params[key] = value
        elif isinstance(value, Module):
            self._modules[key] = value
        super(Module, self).__setattr__(key, value)

    def __call__(self, *inputs: List[Tensor]) -> Tensor:
        return self.forward(*inputs)


class Linear(Module):
    def __init__(
        self,
        input_size: int,
        output_size: int,
        initializer: Callable = np.random.random,
        bias: bool = True,
    ):
        """A Linear module computes defines weights W, optionally biases b, and
        computers wX + b.

        Weight vector will have shape (input size, output size)

        Args:
            input_size: dimension of input vectors
            output_size: dimension of output vectors
            initializer: how to initialize weights and biases
            bias: whether or not to include the bias term; not needed for, e.g. embeddings
        """
        super(Linear, self).__init__()
        self.weights = Tensor(initializer((input_size, output_size)), name="W")
        self.bias = bias
        if self.bias:
            self.biases = Tensor(initializer((1, output_size)), name="b")

    def forward(self, inputs: Tensor):
        mul_node = ops.matmul(inputs, self.weights)
        if self.bias:
            # NOTE: this is a hack-ish way of handling shape issues with biases
            expanded_biases = ops.copy_rows(self.biases, num=inputs.value.shape[0])
            return ops.add(mul_node, expanded_biases)
        return mul_node



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="edugrad",
    version="0.0.6",
    author="Shane Steinert-Threlkeld",
    author_email="ssshanest@gmail.com",
    description="Basic computation graph, for pedagogical purposes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shanest/edugrad",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.17",
        "networkx>=2.4"
    ]
)


""" Copyright 2020-2021 Shane Steinert-Threlkeld

    This file is part of edugrad.

    edugrad is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    edugrad is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with edugrad.  If not, see <https://www.gnu.org/licenses/>.
"""

"""This module defines the Tensor class, which is a wrappers on numpy arrays.

Typically, you will never manually create Tensors.  They are created from 
numpy arrays in two places:
    * in DataIterators (see `data.py`)
    * via `tensor_ops` (see `ops.py`)
"""

from typing import Iterable
import numpy as np
import networkx as nx
import edugrad.ops as ops


class Tensor:
    """Main Tensor class, wrapping a numpy array with gradients etc.

    Attributes:
        value: numpy array with the tensor's value
        parents: Iterable of Tensors, this node's parents in a dynamic graph
        name: a string to name the Tensor
        grad: gradient, if it has been computed; numpy array of same shape as value
    """

    def __init__(self, value: np.ndarray, parents: Iterable = (), name: str = None):
        """Initialize values, set gradients to zero. """
        self.value = value
        self.parents = parents
        self.name = name
        self.grad = np.zeros(value.shape)

    def backward(self) -> None:
        """Run backward pass from a scalar tensor.

        All Tensors in the graph above this one will wind up having their
        gradients stored in `grad`.

        Raises:
            ValueError, if this is not a scalar.
        """
        if not np.isscalar(self.value):
            raise ValueError("Can only call backward() on scalar Tensors.")
        # dL / dL = 1
        self.grad = np.ones(self.value.shape)
        # NOTE: building a graph, then sorting, is not maximally efficient
        # but the graph can be used for visualization etc
        graph = self.get_graph_above()
        reverse_topological = reversed(list(nx.topological_sort(graph)))
        for tensor in reverse_topological:
            tensor._backward()

    def _backward(self) -> None:
        """This is a private helper method.

        Computes the upstream gradients and populates the parent Tensors.
        Pass is the default behavior, which will be evoked only for leaf nodes,
        i.e. those with no parents.

        It gets populated by `ops.tensor_op()`
        """
        pass

    def get_graph_above(self) -> nx.DiGraph:
        """Get the full computation graph of Tensors above the present one. """
        graph = nx.DiGraph()
        visited = set()

        def visit(value: Tensor):
            if value not in visited:
                for parent in value.parents:
                    graph.add_edge(parent, value)
                    visit(parent)
            visited.add(value)

        visit(self)
        return graph

    def __add__(self, other):
        return ops.add(self, other)

    def __radd__(self, other):
        return ops.add(other, self)

    def __sub__(self, other):
        return ops.minus(self, other)

    def __pow__(self, other):
        return ops.power(self, exponent=other)